class BadUML(Exception):
    """Raised in case UML design is incorrect."""

class BadReference(BadUML):
    """
    Raised in case UML object Reference references non-existant operation 
    in other UML Class.
    """
    def __init__(self, tname, oper):
        self.tname = tname
        self.oper = oper

    def __str__(self):
        return ("UML Class '%s' has no operation '%s' to use as reference. " 
                "You must define valid roles for both end "
                "points." % (self.tname, self.oper))

class BadOperation(BadUML):
    """Raised when a operation is placed at a wrong place."""

class IncompleteReference(BadUML):
    """
    Raised in case UML object Reference is not connected to both end points.
    """
    def __init__(self, connected):
        self.connected = connected

    def __str__(self):
        return ("One end point of an UML Reference is disconnected, "
                "and the other end is connected to the UML Class "
                "'%s'." % self.connected)

class ReferenceDuplicate(BadUML):
    """Raised when finding the same reference between same UML Class."""
    def __init__(self, dtable, dcolumn):
        self.dtable = dtable
        self.dcolumn = dcolumn

    def __str__(self):
        return ("Found a reference duplicate at UML Class '%s' involving "
                "attribute '%s'." % (self.dtable, self.dcolumn))

class CyclicReference(BadUML):
    """Raise when finding a cyclic reference in an UML Class."""
    def __init__(self, ctable, ccolumn):
        self.ctable = ctable
        self.ccolumn = ccolumn

    def __str__(self):
        return ("Found a cyclic reference at UML Class '%s' involving "
                "attribute '%s'." % (self.ctable, self.ccolumn))

class NoColumnType(Warning):
    """Raise when a attribute has no type defined."""

class DisconnectedObject(Warning):
    """Raised when both end points of an UML object are disconnected."""

class BaseParser(object):
    """
    UML parsers should inherit from this class and override methods 
    not implemented.
    """

    def __init__(self, uml_class, uml_inheritance, uml_association, 
                 uml_root, nstracker):
        self.dbname = None
        self.uml_layer = None

        self.root = uml_root
        self.nstracker = nstracker
        self.uml_read = {uml_class: self.parse_table_from_uml_class,
                         uml_inheritance: self.parse_table_inheritance,
                         uml_association: self.parse_table_association}

    def parse_uml(self):
        raise NotImplementedError
    
    def parse_table_from_uml_class(self):
        raise NotImplementedError

    def parse_table_inheritance(self):
        raise NotImplementedError

    def parse_table_association(self):
        raise NotImplementedError



"""
Storm ORM code generator.
"""

from uml2orm.codegen.orm import PYTHON_MAPPING

INDENT = " " * 4
PREFIX = "sl"
STORM_MAPPING = {"integer": "Int", "int": "Int",
                 
                 "text": "Unicode", "varchar": "Unicode", "string": "Unicode", 
                 "str": "Unicode", "char": "Unicode",
                 
                 "date": "DateTime", "datetime": "DateTime", 
                 "timestamp": "DateTime",
                  
                 "bool": "Bool",

                 "enum": "Enum"}
STORM_IMPORT = "import storm.locals as %s" % PREFIX

TMPL_CLASS = """class %(name)s(%(inheritance)s):
%(attributes)s
   
%(indent)sdef __init__(self, %(args)s):
%(attribs_init)s"""

class StormConstructor(object):
    """Creates ORM code for Storm."""
    
    def __init__(self, database):
        self.database = database
        self.dtables = self.database.tables
        
    def create_orm(self):
        """Creates all ORM code."""
        # missing code generation for controlling the database, and i am
        # not sure if it is a good thing to generate this kind of code.
        
        return self.create_base_orm()
        
    def create_base_orm(self):
        """Creates ORM code that maps database tables to classes."""
        default_inheritance = '%s.Storm' % PREFIX
        classes = []
        extra_imports = []
        
        for table in self.dtables.values():
            new_class = {'name': self._class_name(table.name), 
                         'inheritance': table.inherits, 'attributes': None, 
                         'indent': INDENT, 'args': None, 'attribs_init': None}

            # setup class inheritance
            inherits = []
            for inherit in new_class['inheritance']:
                inherits.append(self._class_name(self.dtables[inherit].name))
            new_class['inheritance'] = ', '.join(inherits)
            if not new_class['inheritance']:
                new_class['inheritance'] = default_inheritance
                
            # setup class attributes
            columns = []
            references = []
            for column in table.columns:
                column_dtype = column.dtype.split()
                dtype = self._mapping(column_dtype[0])
                if dtype == -1: # mapping not found
                    raise KeyError(("Mapping '%s' specified at UML Class "
                                    "'%s' is invalid" % (column_dtype[0],
                                                         table.name)))
                
                if column.operation:
                    if column.operation[0] == 'pk':
                        dtype += '(primary=True)'
                    elif column.operation[0] == 'fk':
                        dtype += '()'
                        tref = table.references[column.name]
                        references.append((column.name, PREFIX, column.name,
                                self._class_name(self.dtables[tref[0]].name), 
                                tref[1]))
                                
                else:
                    dtype += '()'
                    
                columns.append((column.name, dtype, column.defval,
                                'notnull' in column_dtype))

            # preparing for TMPL_CLASS
            attribs = []
            if table.table:
                attribs.append("%s__storm_table__ = \"%s\"" % (INDENT, 
                                                        table.name.lower()))
            args = []
            args_null = []
            attribs_init = []
            for col in columns:
                attribs.append('%s%s = %s.%s' % (INDENT, col[0], PREFIX, 
                                                 col[1]))
                
                if col[2]: # default value set
                    imports, python_code = PYTHON_MAPPING[col[2].lower()]
                    extra_imports.append(imports)
                    
                    attribs_init.append("%sif not self.%s:" % (INDENT * 2, 
                                                               col[0]))
                    attribs_init.append("%sself.%s = %s" % (INDENT * 3, col[0],
                                                            python_code))
                    attribs_init.append("%selse:" % (INDENT * 2))
                    attribs_init.append("%sself.%s = %s" % (INDENT * 3, col[0],
                                                            col[0]))
                else:
                    attribs_init.append('%sself.%s = %s' % (INDENT * 2, col[0], 
                                                            col[0]))
                if col[3]:
                    args.append(col[0])
                else:
                    args_null.append("%s=None" % col[0])
            
            for ref in references:
                if ref[0].startswith("fk_"):
                    refname = ref[0][3:]
                else:
                    refname = "fk_%s" % ref[0]
                    
                attribs.append('%s%s = %s.Reference(%s, "%s.%s")' % (INDENT,
                                                       refname, ref[1], ref[2], 
                                                       ref[3], ref[4]))
            
            args.extend(args_null)
            
            new_class['attributes'] = '\n'.join(attribs)
            new_class['args'] = ', '.join(args)
            new_class['attribs_init'] = '\n'.join(attribs_init)
            
            classes.append(TMPL_CLASS % new_class)

        extra_imports = ['import %s' % eimp for eimp in extra_imports]
        extra_imports = '\n'.join(extra_imports)
        return extra_imports, STORM_IMPORT, '\n\n'.join(classes)
    
    def _mapping(self, tomap):
        """Sanitize tomap and then use it at STORM_MAPPING."""
        parenthesis = tomap.find('(') 
        if parenthesis != -1: # sanitizing "some(thing)"
            tomap = tomap[:parenthesis]
        
        if tomap not in STORM_MAPPING:
            return -1
        
        return STORM_MAPPING[tomap]
    
    def _class_name(self, cname):
        """Returns an appropriate name for a class."""
        return cname.replace('_', ' ').title().replace(' ', '')


"""
Generates sqlite SQL code based on database objects passed in.
"""

SQL_MAPPING = {'varchar': 'TEXT', 'str': 'TEXT', 'string': 'TEXT',
               'int': 'INTEGER',
               'date': 'TIMESTAMP', 'datetime': 'TIMESTAMP',
               'notnull': 'NOT NULL',
               'pk': 'INTEGER%(null)s PRIMARY KEY',
               'fk': ("INTEGER%(null)s CONSTRAINT %(fk_column)s\n"
               "              REFERENCES %(table)s(%(column)s) %(cascade)s"),
               'cascade': "ON DELETE CASCADE"}

# DELETE trigger templates
TMPL_DEL_TRIGGER_TITLE = "%(refdtable)s_%(kind)s_delete"
TMPL_DEL_TRIGGER_BASE = """CREATE TRIGGER %(title)s
    BEFORE DELETE ON %(reftable)s
    FOR EACH ROW BEGIN
        %(del_oper)s
    END;"""
TMPL_DEL_TRIGGER = """SELECT RAISE(ROLLBACK, "Bad DELETE on table '%(reftable)s', it still" ||
                               "holds references.")
        WHERE (SELECT %(column)s FROM %(table)s 
               WHERE %(column)s = OLD.%(refcolumn)s) IS NOT NULL;"""
TMPL_DEL_TRIGGER_CASCADE = """DELETE FROM %(table)s WHERE %(table)s.%(column)s = OLD.%(refcolumn)s;"""

# INSERT/UPDATE trigger templates
TMPL_IU_TRIGGER_TITLE = "%(table)s_%(lower_op)s_bad_%(column)s"
TMPL_IU_TRIGGER = """CREATE TRIGGER %(title)s
    BEFORE %(operation)s ON %(table)s
    FOR EACH ROW BEGIN
        SELECT RAISE(ROLLBACK, "Bad %(operation)s on table '%(table)s', invalid" ||
                               "%(column)s especified.")
        WHERE%(null)s (SELECT %(refcolumn)s FROM %(reftable)s 
                       WHERE %(refcolumn)s = NEW.%(column)s) IS NULL;
    END;"""
TMPL_IU_TRIGGER_NULL = "NEW.%s IS NOT NULL AND"

# INDEX template 
TMPL_INDEX = "CREATE%(unique)s INDEX %(index_name)s ON %(table)s(%(column)s);"

# table base structure
BASETABLE = "CREATE TABLE %(name)s (\n%(columns)s\n);"

class SqliteConstructor(object):
    """Construct complete sqlite code based on the database content."""

    def __init__(self, database):
        self.database = database
        self.dtables = self.database.tables
        self.trigger_names = [ ]
        self.trigger_prefix = { } # used to ensure uniqueness in trigger names
        self.index_names = { } # used to ensure uniqueness in index names

    def create_sql(self):
        """
        Construct complete sqlite code based on the self.database 
        content. Returns (in the following order): Tables SQL, 
        Triggers SQL, Indexes SQL, Drop tables SQL, Drop triggers SQL, 
        Drop indexes SQL.
        """
        tables = { } # table_name: {column1, column2, .., columnN}
        sql_triggers = [ ] # triggers code
        sql_indexes = [ ] # indexes code

        for table in self.dtables.values():
            if not table.table:
                # Table only inherited by others (probably), no SQL code 
                # gets produced for this.
                continue
            
            tname = table.name.lower()
            tables[tname], indexes, triggers = self._build_table(table)
            sql_indexes.extend(indexes)
            sql_triggers.extend(triggers)

        # format sql code nicely =)
        sql_tables = [ ]
        for tname, columns in tables.items():
            cols = ["%12s  %s," % (col[0], col[1]) for col in columns]
            table_dict = {"name": tname,
                          # remove last comma from table columns
                          "columns": '\n'.join(cols)[:-1]}

            sql_tables.append(BASETABLE % table_dict)
 
        sql_tables = '\n\n'.join(sql_tables)
        sql_triggers = '\n\n'.join(sql_triggers)
        sql_indexes = '\n'.join(sql_indexes)
        sql_del_tables = '\n'.join("DROP TABLE %s;" % t for t in tables.keys())
        sql_del_triggers = '\n'.join("DROP TRIGGER %s;" % t 
                                        for t in self.trigger_names)
        sql_del_indexes = '\n'.join("DROP INDEX %s;" % i 
                                        for i in self.index_names)

        return (sql_tables, sql_triggers, sql_indexes, sql_del_tables, 
                sql_del_triggers, sql_del_indexes)

    def _build_table(self, wtable, calling_from=None):
        """
        Build all the columns for wtable (working table, or current table).
        Lookups for inheritances and construct columns for them first.
        It will also construct triggers needed for this table aswell.
        """
        new_columns, new_triggers, new_indexes = [ ], [ ], [ ]
        wtable_name = wtable.name.lower() # using lowercase for table name

        # construct code from inherited tables first
        for inherit in wtable.inherits:
            columns, indexes, triggers = self._build_table(
                                            self.dtables[inherit], wtable_name)
            new_columns.extend(columns)
            new_indexes.extend(indexes)
            new_triggers.extend(triggers)

        # indexes code
        if wtable.indexes:
            index_name = self._good_name("%s_indx" % '_'.join(wtable.indexes), 
                                         self.index_names)
            new_indexes.append(TMPL_INDEX % {
                'index_name': index_name,
                'table': calling_from if not wtable.table else wtable_name,
                'column': ', '.join(wtable.indexes), 'unique': ''})#'UNIQUE' if indx.unique else ''})

        # columns code
        for column in wtable.columns:
            column_dtype = column.dtype.split()

            if column.operation and column.operation[0] in SQL_MAPPING:
                column_subs = {'null': '', 'fk_column': None, 'table': None,
                               'column': None, 'cascade': ''}
                if 'notnull' in column_dtype:
                    column_subs['null'] = ' %s' % SQL_MAPPING['notnull']

                if column.name in wtable.references:
                    # need to create triggers aswell
                    ret = self._build_triggers(wtable_name, column.name, 
                                               wtable.references[column.name],
                                               column_dtype, 
                                               column.operation[1])
                    new_triggers.extend(ret)
                
                    ref = wtable.references[column.name]
                    column_subs['fk_column'] = column.name
                    column_subs['table'] = self.dtables[ref[0]].name
                    column_subs['column'] = ref[1]
                    if column.operation[1] == 'cascade':
                        column_subs['cascade'] = SQL_MAPPING['cascade']

                newc = SQL_MAPPING[column.operation[0]] % column_subs
        
            else:
                if not column_dtype:
                    # set TEXT as a fallback
                    column_dtype.append(SQL_MAPPING['str'])

                newc = ' '.join(SQL_MAPPING[dtype] if dtype in SQL_MAPPING 
                                  else dtype.upper() for dtype in column_dtype)

            new_columns.append((column.name, newc))

        return new_columns, new_indexes, new_triggers


    def _build_triggers(self, wtname, wtcolumn, wcolumnref, column_type, 
                        extra):
        """
        Creates a DELETE trigger for the table being referenced and
        creates triggers for INSERT and UPDATE statements for the table 
        referencing, so data integrity is enforced in sqlite.
        """
        ref = wcolumnref
        tref = self.dtables[ref[0]].name # referenced table name
        tcolumn = ref[1] # referenced table column name

        ddata = {'table': wtname, 'column': wtcolumn,
                 'refcolumn': tcolumn, 'reftable': tref,
                 'refdtable': self._good_name(tref, self.trigger_prefix),
                 'operation': None, 'lower_op': None, 'null': '',
                 'kind': 'bad'}

        sql_triggers = [ ]

        # create triggers for INSERT and UPDATE
        operations = ("INSERT", "UPDATE") 
        if "notnull" not in column_type:
            ddata['null'] = ' %s' % (TMPL_IU_TRIGGER_NULL % wtcolumn)

        for oper in operations:
            ddata['operation'] = oper
            ddata['lower_op'] = ddata['operation'].lower()
            ddata['title'] = TMPL_IU_TRIGGER_TITLE % ddata

            self.trigger_names.append(ddata['title'])
            sql_triggers.append(TMPL_IU_TRIGGER % ddata)
   
        # create trigger for DELETE
        if extra == 'cascade':
            del_oper = TMPL_DEL_TRIGGER_CASCADE % ddata
            ddata['kind'] = 'cascade'
        else:
            del_oper = TMPL_DEL_TRIGGER % ddata

        ddata['del_oper'] = del_oper
        ddata['title'] = TMPL_DEL_TRIGGER_TITLE % ddata
        self.trigger_names.append(ddata['title'])
        sql_triggers.append(TMPL_DEL_TRIGGER_BASE % ddata)

        return sql_triggers


    def _good_name(self, index_name, look_at):
        """Return a unique name based on index_name and keys at look_at."""
        index_name = index_name.lower()
        
        if index_name in look_at:
            look_at[index_name] += 1
            index_name = index_name + str(look_at[index_name])
        else:
            look_at[index_name] = 0

        return index_name



from uml2orm.umlparser import UMLParser
from uml2orm.codegen import create_db_objs

# only sqlite sql construction is supported for now
from uml2orm.codegen.sql.sqlite import SqliteConstructor
# only storm orm construction is supported for now
from uml2orm.codegen.orm.storm import StormConstructor

__version__ = "0.2"

def construct_database(uml_file):
    """Constructs SQL code and ORM code based on an UML description."""
    umlparser = UMLParser(uml_file)
    umlparser.parse_uml()
    
    database_obj = create_db_objs(umlparser.dbname, umlparser.tables)

    sql = SqliteConstructor(database_obj)
    sql = sql.create_sql()
    
    orm = StormConstructor(database_obj)
    orm = orm.create_orm()
    
    return database_obj, sql, orm



"""
Totally incomplete for now.
"""

from uml2orm.umlparser.base import BaseParser

class XmiParser(BaseParser):
    """Handles XMI files."""
    xmi_layer = "XMI.content"
    uml_class = "Class"
    uml_association = "Association"
    
    def __init__(self, xmi_root, nstracker):
        BaseParser.__init__(self, self.uml_class, self.uml_association,
                            xmi_root, nstracker)

        self.find_uml_layer()

    def find_uml_layer(self):
        """Find where UML objects are in XMI file."""
        for child in self.root.getchildren():
            if child.tag != self.xmi_layer:
                #self.uml_layer = child
                continue

            for c in child.getchildren()[0].getchildren()[0].getchildren():
                local = self.nstracker.local_name(c.tag)
                if local == "Model":
                    if len(c.getchildren()) > 1:
                        self.uml_layer = c.getchildren()[0]
                        return

    def build_tables(self):
        """Build tables based on UML."""
        for uml_object in self.uml_layer.getchildren():
            obj_type = self.nstracker.local_name(uml_object.tag)
    
            if obj_type in self.uml_read:
                #getattr(self, self.uml_read[obj_type])(uml_object)
                self.uml_read[obj_type](uml_object)

    def _build_table_from_uml_class(self, uml_obj):
        new_table = Table()
        
        obj_id = uml_obj.attrib['xmi.id']
        new_table.tid = obj_id
        new_table.name = uml_obj.attrib['name']

        for columns in uml_obj.getchildren():
            t_columns = { } 
            for column in columns.getchildren():
                column_name = column.attrib['name']
            """
            if attr.attrib['name'] == 'name':
                new_table.name = self.get_obj_text(attr.getchildren()[0])
                continue

            elif attr.attrib['name'] != 'attributes':
                continue

            t_columns = { } # table columns, 'column name': 'data type'
            for columns in attr.getchildren():
                column_name = None
                for col in columns.getchildren():
                    if col.attrib['name'] == 'name':
                        # column name
                        column_name = self.get_obj_text(col.getchildren()[0])

                    elif col.attrib['name'] == 'type':
                        # column data type
                        data_type = self.get_obj_text(col.getchildren()[0])
                        t_columns[column_name] = data_type
            """

        #new_table.columns.update(t_columns)
        self.tables[obj_id] = new_table

    def _read_uml_association(self, uml_obj): 
        pass


import gzip
from xml.etree.ElementTree import ElementTree, XMLTreeBuilder
from uml2orm.umlparser.dia import DiaParser
from uml2orm.umlparser.xmi import XmiParser

class NamespaceTracker(XMLTreeBuilder):
    def __init__(self):
        XMLTreeBuilder.__init__(self)

        self._parser.StartNamespaceDeclHandler = self._start_ns
        self.namespaces = { }

    def local_name(self, name):
        return self.analyze_name(name)[1]

    def analyze_name(self, name):
        if name[0] == '{':
            ns, local = name[1:].split("}")
        else:
            return None, None, None

        prefix = self.namespaces[ns]
        if prefix is None:
            prefix = u"!Unknown"

        return prefix, local, ns

    def _start_ns(self, prefix, ns):
        self.namespaces[ns] = prefix
        

class UMLParser(object):
    """Opens and parses:
        DIA file (compressed and uncompressed);
        XMI file."""

    def __init__(self, uml_file):
        self.root = None
        self.nstracker = NamespaceTracker()
        
        self.parse(uml_file)
        self.set_appropriate_parser()
        
    def __getattribute__(self, attr):
        try:
            return object.__getattribute__(self, attr)
        except AttributeError:
            return object.__getattribute__(self.parser, attr)

    def parse(self, ufile):
        """Opens uml file and parse it, then store its root."""
        uml_file = gzip.open(ufile)
        etree = ElementTree()

        try:
            self.root = etree.parse(uml_file, self.nstracker)

        except IOError:
            # This probably was caused because we opened the file in binary
            # mode to read a compressed DIA file but the file is not in this
            # format, so it failed. Open it as a normal file now and try to
            # parse it again.
            uml_file = open(ufile)
            self.root = etree.parse(uml_file, self.nstracker)

        finally:
            uml_file.close()

    def set_appropriate_parser(self):
        """Set a parser based on namespace found during ElementTree parsing."""
        uml_ns = self.nstracker.namespaces.values()[0].lower()
        if uml_ns == "dia":
            self.parser = DiaParser
        elif uml_ns == "uml":
            self.parser = XmiParser

    def _get_parser(self):
        """Get parser being used."""
        return self.__parser

    def _set_parser(self, parser):
        """Instanciate an UML parser."""     
        self.__parser = parser(self.root, self.nstracker)


    # Properties
    parser = property(_get_parser, _set_parser)


"""
A script that uses uml2orm for performing code generation.
"""

import sys
import uml2orm

def separator():
    print '*' * 40

if __name__ == "__main__":
    file_input = None
    if len(sys.argv) < 2:
        file_input = raw_input("Especify a filename please: ")
    else:
        file_input = sys.argv[1]

    db_obj, sql, orm = uml2orm.construct_database(file_input)

    print "Database '%s'" % db_obj.name
    separator()
  
    # sql return (indexes):
    #   0 = tables code, 1 = triggers code, 2 = indexes code, 
    #   3 = tables drop code, 4 = triggers drop code, 5 = indexes drop code
    for sql_output in sql:
        if not sql_output:
            continue
        
        print sql_output
    separator()

    # orm return (indexes):
    #   0 = extra imports needed bases on orm code, 1 = orm import,
    #   2 = orm code
    for i, orm_output in enumerate(orm):
        if not orm_output:
            continue

        if i == 2:
            print
        print orm_output



from uml2orm.umlparser.base import BadOperation

class Database(object):
    """Representation of a database."""
    
    def __init__(self, name, tables=None):#, indexes=None):
        self.name = name
        self.tables = tables


class Table(object):
    """Representation of a Database table."""
    
    def __init__(self, tid, name, table, columns=None, references=None, 
                 inherits=None, indexes=None, *args, **kwargs):
        self.table = table # this is false when an UML class is inherited only
        self.tid = tid # table id
        self.name = name # table name
        self.columns = columns # a list of Column instances
        self.references = references # column references (foreign keys)
        self.inherits = inherits # table inheritance
        self.indexes = indexes # columns indexes


class Column(object):
    """Representation of a Table column."""

    def __init__(self, name, dtype, defval, operation=None):
        self.name = name
        self.dtype = dtype # data type
        self.defval = defval # default value
        self.operation = operation


class Index(object):
    """Representation of an Index in a table column."""
    
    def __init__(self, name, table, column):
        self.name = name
        self.table = table
        self.column = column


def create_db_objs(dbname, tables_dict):
    """
    Create Database objects based on the tables dict that were built at an
    UML Reader. No error checking is done here, it is supposed that the
    uml parser that generated tables_dict did all the necessary checkings
    already.
    """
    tables = { }

    for tid, table in tables_dict.items():
        new_table = table
        new_table.update({'tid': tid})
        tcolumns = new_table.pop('columns')
        operations = new_table.pop('operations')
        new_table['columns'] = [0] * len(tcolumns)

        # create Columns and Index objects for current table
        for column, data in tcolumns.items():
            dtype, defval, index = data
            
            if column not in operations:
                new_table['columns'][index] = Column(column, dtype, 
                                                     defval, None)
                continue
            
            opers = operations.pop(column)
            if len(opers) == 1 and opers[0][0] == 'index':
                new_table['indexes'].append(column)
                new_table['columns'][index] = Column(column, dtype, 
                                                     defval, None)
                continue
                
            for op_name, op_type in opers:
                if op_name == "index":
                    new_table['indexes'].append(column)
                    continue
                
                new_table['columns'][index] = Column(column, dtype, defval,
                                                     (op_name, op_type))

        if operations:
            # some invalid operation was created, since it didn't match any
            # attribute in current uml class
            raise BadOperation(("The following operations: '%s' couldn't be "
                                "found at UML Class '%s'"
                                "." % ("', '".join(operations.keys()),
                                       new_table['name'])))

        tables[tid] = Table(**new_table)
    
    return Database(dbname, tables)


from distutils.core import setup
import uml2orm

base_package = "uml2orm"
codegen = "%s/codegen" % base_package
umlparser = "%s/umlparser" % base_package

setup(
    name = base_package,
    version = uml2orm.__version__,
    description = 'uml2orm',
    author='Guilherme Polo',
    author_email = 'ggpolo@gmail.com',
    packages = [base_package, codegen, umlparser, "%s/sql" % codegen,
                "%s/orm" % codegen]            
    )



"""
Maybe all this huge amount of error checking should be placed somewhere else
since all the parsers will need to do this same thing.
"""

import warnings
from uml2orm.consts import DEFAULT_DB_NAME
from uml2orm.umlparser.base import BaseParser
from uml2orm.umlparser.base import (BadUML, BadReference, 
                                    IncompleteReference, ReferenceDuplicate,
                                    CyclicReference)
from uml2orm.umlparser.base import NoColumnType, DisconnectedObject

class DiaParser(BaseParser):
    """Handles DIA files."""
    dia_layer = "layer"
    uml_class = "UML - Class"
    uml_inheritance = "UML - Generalization"
    uml_association = "UML - Association"

    def __init__(self, dia_root, nstracker):
        BaseParser.__init__(self, self.uml_class, self.uml_inheritance,
                            self.uml_association, dia_root, nstracker)

        self.tables = { }
        self.dbname = None
        self.find_uml_layer()


    def get_obj_text(self, dia_str_object):
        """Removes '#' from string object and return the new string."""
        return dia_str_object.text[1:-1]  
    
    
    def find_uml_layer(self):
        """Find where UML objects are in DIA file."""
        for child in self.root.getchildren():
            local = self.nstracker.local_name(child.tag)

            if local == self.dia_layer:
                self.uml_layer = child
                break      


    def parse_uml(self):
        """Build a database based on UML description."""
        inheritance_obj = []
        reference_obj = []

        # First we will parse the tables and store other objects we
        # will need later
        for uml_object in self.uml_layer.getchildren():
            obj_type = uml_object.attrib['type']
    
            if obj_type in self.uml_read:
                if obj_type == self.uml_class:
                    # UML Class (Tables in database)
                    ret = self.uml_read[obj_type](uml_object)

                    if ret:
                        # found database name
                        self.dbname = ret

                elif obj_type == self.uml_inheritance:
                    # UML Generalization (Inheritance)
                    inheritance_obj.append(uml_object)

                else:
                    # UML Association (Table reference)
                    reference_obj.append(uml_object)
        
        # Parse inheritances now
        for inherit in inheritance_obj:
            self.uml_read[self.uml_inheritance](inherit)

        # And finally, parse references
        for ref in reference_obj:
            self.uml_read[self.uml_association](ref)

        # Check for database name
        if self.dbname is None:
            # no UML Class defined as Database, so lets set the default
            # name to it.
            self.dbname = DEFAULT_DB_NAME
            

    def parse_table_from_uml_class(self, uml_obj):
        """Parse table attributes and operations."""

        def parse_table_attributes(table_attrs):
            """Parse UML Class attributes doing proper error checking."""
            errors = [ ]
            # table columns, 'column name': ('data type', 'default value')
            t_columns = { }
            for indx, columns in enumerate(table_attrs):
                t_col = { }

                for col in columns.getchildren():
                    col_n = col.attrib['name']

                    if col_n in ['name', 'type', 'value']:
                        t_col[col_n] = self.get_obj_text(col.getchildren()[0])

                if not t_col['name']: # obrigatory field not filled
                    return None, 1

                elif t_col['name'] in t_columns:  # column duplicated
                    return t_col['name'], 2

                elif not t_col['type']: 
                    # no type especified, a warning will be raised
                    errors.append(t_col['name'])
                    
                t_columns[t_col['name']] = (t_col['type'], t_col['value'], indx)

            return t_columns, errors

        def parse_table_operations(table_operations):
            """Parse UML Class operations doing proper error checking."""
            t_operations = { }

            for ops in table_operations:
                t_op = { }

                for op in ops.getchildren():
                    op_attr = op.attrib['name']
           
                    if op_attr in ['name', 'type', 'stereotype']:
                        t_op[op_attr] = self.get_obj_text(op.getchildren()[0])

                if not t_op['name'] or not t_op['stereotype']:
                    # both fields are obrigatory
                    return None, 1
              
                if t_op['stereotype'] == 'index':
                    t_op['name'] = t_op['name'].split(',')
                else:
                    t_op['name'] = t_op['name'].split()
                 
                for op_name in t_op['name']:
                    op_name = op_name.strip()
                    
                    if op_name in t_operations:
                        for oper in t_operations[op_name]:
                            if oper[0] == t_op['stereotype'].lower():
                                # found a duplicated operation
                                #return (t_op['stereotype'], t_op['name']), 2
                                return (t_op['stereotype'], op_name), 2
                    
                    t_operations.setdefault(op_name, []).append(
                            (t_op['stereotype'].lower(), t_op['type'].lower()))
           
            return t_operations, 0

        # table skeleton
        new_table = {'columns': None, 'operations': None, 'references': {},
                     'inherits': [], 'indexes': [], 'table': True, 
                     'name': None}
        obj_id = uml_obj.attrib['id']
    
        for attr in uml_obj.getchildren():
            attr_name = attr.attrib['name']

            if attr_name == 'name':
                # Found name for this UML class
                new_table['name'] = self.get_obj_text(attr.getchildren()[0])
           
            # Check if we are dealing with a real database Table
            elif attr_name == 'stereotype':
                stereotype = self.get_obj_text(attr.getchildren()[0]).lower()

                if stereotype == 'database':
                    # This is an UML class that defines a database, not
                    # a table actually.
                    db_name = new_table['name']
                    
                    return db_name

                elif stereotype == 'none':
                    # This UML class will not generate SQL code, it is
                    # just being subclassed by some other table. This is
                    # used only to construct ORM code.
                    new_table['table'] = False

            elif attr_name == 'attributes':
                t_columns, ret = parse_table_attributes(attr.getchildren())
                if ret == 1:
                    raise BadUML(("No name defined for a column at UML Class "
                                  "'%s'." % new_table['name']))
                elif ret == 2:
                    raise BadUML(("Found a duplicated column ('%s') at UML "
                                  "Class "
                                  "'%s'." % (t_columns, new_table['name'])))
                elif ret:
                    for err_col in ret:
                        warnings.warn(("Column '%s' at UML Class '%s' has no "
                                       "type defined, assuming "
                                       "TEXT." % (err_col, new_table['name'])),
                                       NoColumnType)

            elif attr_name == 'operations':
                t_operations, ret = parse_table_operations(attr.getchildren())
                if ret == 1:
                    raise BadUML(("An operation at UML Class '%s' is missing "
                                  "both or one of the following data: "
                                  "name, stereotype." % new_table['name']))
                elif ret == 2:
                    raise BadUML(("Found a duplicated operation %s at UML "
                                  "Class "
                                  "'%s'." % (t_operations, new_table['name'])))

        new_table['columns'] = t_columns
        new_table['operations'] = t_operations 
        self.tables[obj_id] = new_table
    

    def parse_table_inheritance(self, uml_obj):
        """Find and set tables subclasses."""
        for attr in uml_obj.getchildren():
            if self.nstracker.local_name(attr.tag) == "connections":
                inherit_from, inheriter = None, None

                for inherit in attr.getchildren():
                    if inherit.attrib['handle'] == '0':
                        inherit_from = inherit.attrib['to']
                    else:
                        inheriter = inherit.attrib['to']

                conn = None
                if inherit_from is None and inheriter is None:
                    warnings.warn(("Both end points of an UML Generalization "
                                   "are disconnected, discarding it."),
                                  DisconnectedObject)
                    return
                elif inherit_from is None:
                    conn = self.tables[inheriter]['name']
                elif inheriter is None:
                    conn = self.tables[inherit_from]['name']

                if conn:
                    raise BadUML(("An UML Generalization object is connected "
                                  "to '%s' but the other end point is "
                                  "disconnected." % conn))
                    
                self.tables[inheriter]['inherits'].append(inherit_from)


    def parse_table_association(self, uml_obj):
        """Find and set table association based on uml_obj."""
        for attr in uml_obj.getchildren():
            # connections
            if self.nstracker.local_name(attr.tag) == "connections":
                conns = [c.attrib['to'] for c in attr.getchildren()]
                continue

            if attr.attrib['name'] != 'ends':
                continue

            # ends
            roles = [self.get_obj_text(end.getchildren()[0]) 
                     for ends in attr.getchildren() 
                        for end in ends.getchildren() 
                            if end.attrib['name'] == 'role']

        # Check for badly connected references.
        if not len(conns):
            warnings.warn(("Both end points of an UML Reference with roles: "
                           "'%s' are disconnected, discarding "
                           "it. " % ' and '.join(roles)), 
                           DisconnectedObject)
            return

        elif len(conns) == 1:
            raise IncompleteReference(self.tables[conns[0]]['name'])
        
        tables = (self.tables[conns[0]], self.tables[conns[1]])

        # Check for inexistent references
        if not roles[0] in tables[0]['operations']:
            raise BadReference(tables[0]['name'], roles[0])

        if not roles[1] in tables[1]['operations']:
            raise BadReference(tables[1]['name'], roles[1])
        
        # Check for duplicated references
        if roles[0] in tables[0]['references']:
            raise ReferenceDuplicate(tables[0]['name'], roles[0])
        if roles[1] in tables[1]['references']:
            raise ReferenceDuplicate(tables[1]['name'], roles[1])

        # Check for cyclic reference
        if conns[0] == conns[1]:
            raise CyclicReference(tables[0]['name'], roles[0])
      
        # Passed all the tests, it is ok to store the reference now
        # The following for loop and if condition are used to "detect"
        # what UML Class is being referenced and which one is referencing
        for oper in tables[0]['operations'][roles[0]]:
            if oper[0] == 'fk':
                tables[0]['references'][roles[0]] = (conns[1], roles[1])
                return

        # If no 'fk' operation was found at tables[0] it means tables[1]
        # is the one referencing and table[0] is referenced.
        tables[1]['references'][roles[1]] = (conns[0], roles[0])
